#define _GNU_SOURCE

#include <stdio.h>
#include <unistd.h>
#include <pwd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#if defined(READLINE)
#include <readline/readline.h>
#elif defined(EDITLINE)
#include <histedit.h>
#endif

#define ftl(...) do { fprintf( stderr, __VA_ARGS__ ); exit(2); } while (0)

#include "mpw-algorithm.h"
#include "mpw-util.h"

#define MP_env_fullname     "MP_FULLNAME"
#define MP_env_sitetype     "MP_SITETYPE"
#define MP_env_sitecounter  "MP_SITECOUNTER"
#define MP_env_algorithm    "MP_ALGORITHM"

static void usage() {

    fprintf( stderr, "Usage: mpw [-u name] [-c counter] site\n\n" );
    fprintf( stderr, "    -u name      Specify the full name of the user.\n"
            "                 Defaults to %s in env.\n\n", MP_env_fullname );
    fprintf( stderr, "    -c counter   The value of the counter.\n"
            "                 Defaults to %s in env or 1.\n\n", MP_env_sitecounter );
    fprintf( stderr, "    -V version   The algorithm version to use.\n"
            "                 Defaults to %s in env or %d.\n\n", MP_env_algorithm, MPAlgorithmVersionCurrent );
    fprintf( stderr, "    ENVIRONMENT\n\n"
            "        MP_FULLNAME    | The full name of the user.\n"
            "        MP_SITETYPE    | The default password template.\n"
            "        MP_SITECOUNTER | The default counter value.\n\n" );
    exit( 0 );
}

static char *homedir(const char *filename) {

    char *homedir = NULL;
    struct passwd *passwd = getpwuid( getuid() );
    if (passwd)
        homedir = passwd->pw_dir;
    if (!homedir)
        homedir = getenv( "HOME" );
    if (!homedir)
        homedir = getcwd( NULL, 0 );

    char *homefile = NULL;
    asprintf( &homefile, "%s/%s", homedir, filename );
    return homefile;
}

static char *getlinep(const char *prompt) {

    char *buf = NULL;
    size_t bufSize = 0;
    ssize_t lineSize;
    fprintf( stderr, "%s", prompt );
    fprintf( stderr, " " );
    if ((lineSize = getline( &buf, &bufSize, stdin )) < 0) {
        free( buf );
        return NULL;
    }
    buf[lineSize - 1] = 0;
    return buf;
}

int main(int argc, char *const argv[]) {

    // Read the environment.
    char *fullName = getenv( MP_env_fullname );
    const char *masterPassword = NULL;
    const char *siteName = NULL;
    uint32_t siteCounter = 1;
    const char *siteCounterString = getenv( MP_env_sitecounter );
    MPAlgorithmVersion algorithmVersion = MPAlgorithmVersionCurrent;
    const char *algorithmVersionString = getenv( MP_env_algorithm );
    if (algorithmVersionString && strlen( algorithmVersionString ))
        if (sscanf( algorithmVersionString, "%u", &algorithmVersion ) != 1)
            ftl( "Invalid %s: %s\n", MP_env_algorithm, algorithmVersionString );

    // Read the options.
    for (int opt; (opt = getopt( argc, argv, "u:c:V:h" )) != -1;)
        switch (opt) {
            case 'u':
                fullName = optarg;
                break;
            case 'c':
                siteCounterString = optarg;
                break;
            case 'V':
                if (sscanf( optarg, "%u", &algorithmVersion ) != 1)
                    ftl( "Not a version: %s\n", optarg );
                break;
            case 'h':
                usage();
                break;
            default:
                ftl("Unexpected option: %c", opt);
        }
    if (optind < argc)
        siteName = argv[optind];

    // Convert and validate input.
    if (!fullName && !(fullName = getlinep( "Your full name:" )))
        ftl( "Missing full name.\n" );
    if (!siteName && !(siteName = getlinep( "Site name:" )))
        ftl( "Missing site name.\n" );
    if (siteCounterString)
        siteCounter = (uint32_t)atol( siteCounterString );
    if (siteCounter < 1)
        ftl( "Invalid site counter: %d\n", siteCounter );
    trc( "algorithmVersion: %u\n", algorithmVersion );

    // Read the master password.
    char *mpwConfigPath = homedir( ".mpw" );
    if (!mpwConfigPath)
        ftl( "Couldn't resolve path for configuration file: %d\n", errno );
    trc( "mpwConfigPath: %s\n", mpwConfigPath );
    FILE *mpwConfig = fopen( mpwConfigPath, "r" );
    free( mpwConfigPath );
    if (mpwConfig) {
        char *line = NULL;
        size_t linecap = 0;
        while (getline( &line, &linecap, mpwConfig ) > 0) {
            char *lineData = line;
            if (strcmp( strsep( &lineData, ":" ), fullName ) == 0) {
                masterPassword = strcpy( malloc( strlen( lineData ) ), strsep( &lineData, "\n" ) );
                break;
            }
        }
        mpw_free( line, linecap );
    }
    while (!masterPassword || !strlen(masterPassword))
        masterPassword = getpass( "Your master password: " );

    // Summarize operation.
    const char* identIcon = mpw_identicon( fullName, masterPassword );
    //fprintf( stderr, "%s's password for %s:\n[ %s ]: ", fullName, siteName, identIcon );

    // Output the password.
    const uint8_t *masterKey = mpw_masterKeyForUser(
            fullName, masterPassword, algorithmVersion );
    mpw_freeString( masterPassword );
    if (!masterKey)
        ftl( "Couldn't derive master key." );

    MPSiteType siteType = MPSiteTypeGeneratedLong;
    MPSiteVariant siteVariant = MPSiteVariantPassword;
    const char *siteContextString = NULL;
    const char *sitePassword = mpw_passwordForSite(
            masterKey, siteName, siteType, siteCounter, siteVariant, siteContextString, algorithmVersion );
    if (!sitePassword)
    {
        mpw_free( masterKey, MP_dkLen );
        ftl( "Couldn't derive site password." );
    }

    fprintf( stdout, "%-12s %s\n\n", "IdentIcon", identIcon );
    char *typeNames[] = { "Maximum", "Long", "Medium", "Basic", "Short", "PIN", "Name", "Phrase" };
    for( int i=0; i<sizeof(typeNames)/sizeof(char *); i++)
    {
      siteType = mpw_typeWithName( typeNames[i] );
      sitePassword = mpw_passwordForSite( masterKey, siteName, siteType, siteCounter, siteVariant, siteContextString, algorithmVersion );
      fprintf( stdout, "%-12s %s\n", typeNames[i], sitePassword );
    }
    mpw_free( masterKey, MP_dkLen );
    return 0;
}
